#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
import sys

import click

import neptune
from neptune.api_exceptions import NotebookNotFound


def sync(project_name, path, update_flag=None, new_flag=None):
    verify_file(path)
    absolute_path = os.path.abspath(path)
    project = neptune.init(project_qualified_name=project_name)
    notebook_json = load_notebook_json(path)
    nbformat = int(notebook_json["nbformat"]) if "nbformat" in notebook_json else 0

    if nbformat < 4:
        show_nbformat_to_old_error()

    if new_flag:
        upload_new_notebook(project, path, notebook_json)
    else:
        notebook_id = get_notebook_id_from_metadata(notebook_json)
        if notebook_id:
            notebook = fetch_notebook(project, notebook_id)
            if notebook:
                old_path = notebook.get_path()
                if absolute_path == old_path or update_flag:
                    upload_new_checkpoint(project, notebook, path)
                else:
                    show_path_changed_error(project_name, old_path, absolute_path, path)
            elif update_flag:
                return show_unknown_notebook_error(project_name, path)
            else:
                upload_new_notebook(project, path, notebook_json)
        else:
            if update_flag:
                return show_unknown_notebook_error(project_name, path)
            else:
                upload_new_notebook(project, path, notebook_json)


def upload_new_notebook(project, path, notebook_json):
    notebook = project.create_notebook()
    save_notebook_metadata(path, notebook_json, notebook.id)
    notebook.add_checkpoint(path)
    print_link_to_notebook(project, notebook.get_name(), notebook.id)


def upload_new_checkpoint(project, notebook, path):
    notebook.add_checkpoint(path)
    print_link_to_notebook(project, notebook.get_name(), notebook.id)


def load_notebook_json(path):
    with open(path) as f:
        return json.load(f)


def get_notebook_id_from_metadata(notebook_json):
    metadata = None
    if 'metadata' in notebook_json:
        metadata = notebook_json['metadata']

    neptune_metadata = None
    if metadata and 'neptune' in metadata:
        neptune_metadata = metadata['neptune']

    notebook_id = None
    if neptune_metadata and 'notebookId' in neptune_metadata:
        notebook_id = neptune_metadata['notebookId']

    return notebook_id


def fetch_notebook(project, notebook_id):
    try:
        return project.get_notebook(notebook_id)
    except NotebookNotFound:
        return None


def save_notebook_metadata(path, notebook_json, notebook_id):
    metadata = {}
    if 'metadata' in notebook_json:
        metadata = notebook_json['metadata']

    metadata['neptune'] = {'notebookId': notebook_id}
    notebook_json['metadata'] = metadata

    with open(path, 'w') as f:
        f.write(json.dumps(notebook_json))


def verify_file(path):
    if not os.path.exists(path):
        click.echo("ERROR: File `{}` doesn't exist".format(path), err=True)
        sys.exit(1)

    if not os.path.isfile(path):
        click.echo("ERROR: `{}` is not a file".format(path), err=True)
        sys.exit(1)

    if not path.endswith(".ipynb"):
        click.echo("ERROR: '{}' is not a correct notebook file. Should end with '.ipynb'.".format(path), err=True)
        sys.exit(1)


def print_link_to_notebook(project, notebook_name, notebook_id):
    try:
        print("{base_url}/{project}/n/{notebook_name}-{notebook_id}".format(
            base_url=project.client.api_address,
            project=project.full_id,
            notebook_name=notebook_name,
            notebook_id=notebook_id
        ))
    except Exception:
        pass


def show_unknown_notebook_error(project_name, path):
    project_option = " --project {}".format(project_name) if project_name else ""
    click.echo("ERROR: Cannot update notebook {} since it is not known to Neptune. "
               "Use following command to create new notebook in Neptune.\n\n"
               "    neptune notebook sync{} {}\n"
               .format(path, project_option, path), err=True)
    sys.exit(1)


def show_path_changed_error(project_name, old_path, new_absolute_path, path):
    project_option = " --project {}".format(project_name) if project_name else ""
    click.echo("ERROR: Notebook path changed since last checkpoint from\n\n"
               "    {}\n\nto\n\n    {}\n\n"
               "Use following command if you want to create new notebook\n\n"
               "    neptune notebook sync --new{} {}\n\n"
               "or following command if you want to update existing notebook\n\n"
               "    neptune notebook sync --update{} {}\n"
               .format(old_path, new_absolute_path, project_option, path, project_option, path), err=True)
    sys.exit(1)


def show_nbformat_to_old_error():
    click.echo("ERROR: Version of this notebook is too old. "
               "Neptune support notebooks in version 4.0 or greater.", err=True)
    sys.exit(1)



# Original source: https://github.com/Changaco/version.py
import os
import re
from distutils.cmd import Command
from subprocess import CalledProcessError, call, check_output

PREFIX = ''
INITIAL_VERSION = '0.0.0'

tag_re = re.compile(r'\btag: %s([0-9][^,]*)\b' % PREFIX)
version_re = re.compile('^Version: (.+)$', re.M)


def get_git_version():
    if "VERSION" in os.environ:
        return os.environ.get("VERSION")

    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search('$Format:%D$')
    if version:
        return version.group(1)

    # Get the version using "git describe".
    try:
        cmd = 'git describe --tags --match %s[0-9.]* --dirty' % PREFIX
        version = check_output(cmd.split()).decode().strip()[len(PREFIX):]
    except CalledProcessError:
        version = INITIAL_VERSION + '+' + check_output('git rev-parse HEAD'.split()).decode().strip()[:10]
        if call('git diff --quiet'.split()) != 0:
            version += '.dirty'

    # PEP 440 compatibility
    if '-' in version:
        version_tokens = version.split('-')
        version = version_tokens[0] + '+' + '.'.join(version_tokens[1:])

    return version


class GitVersion(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        version = get_git_version()
        self.distribution.metadata.version = version

        with open('VERSION', 'w') as version_file:
            version_file.write(version)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from IPython.display import display, JSON
import json

def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': '../packages/nbextension',
        'dest': 'neptune-notebooks',
        'require': 'neptune-notebooks/neptune-notebook'
    }]

# A display class that can be used within a notebook. 
#   from neptune-notebooks import JSON
#   JSON(data)
    
class JSON(JSON):
    """A display class for displaying JSON visualizations in the Jupyter Notebook and IPython kernel.
    
    JSON expects a JSON-able dict, not serialized JSON strings.

    Scalar types (None, number, string) are not allowed, only dict containers.
    """

    def _ipython_display_(self):
        bundle = {
            'application/json': self.data,
            'text/plain': '<neptune-notebooks.JSON object>'
        }
        metadata = {
            'application/json': self.metadata
        }
        display(bundle, metadata=metadata, raw=True) 



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click


@click.group()
def notebook():
    pass


@notebook.command()
@click.argument('path', required=True)
@click.option('--new', is_flag=True, help='Creates a new notebook instead of updating existing one.')
@click.option('--update', is_flag=True,
              help='Indicates that an existing notebook should be updated if the path was changed.')
@click.option('--project', '-p', help='Project name')
def sync(project, update, new, path):
    # We do not want to import anything if process was executed for autocompletion purposes.
    import sys
    from neptune_notebooks.sync import sync as run_sync

    if new and update:
        click.echo("ERROR: --new and --update flags are mutually exclusive.")
        sys.exit(1)

    return run_sync(project_name=project, path=path, update_flag=update, new_flag=new)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from mock import MagicMock, patch

from neptune.api_exceptions import NotebookNotFound

from neptune_notebooks.sync import sync


class TestSync(unittest.TestCase):

    NOTEBOOK_JSON = {'nbformat': 4, 'metadata': {'neptune': {'notebookId': '6e1cf9fc-723f-11e9-b34f-1ff6c9882fdc'}}}
    UNKNOWN_NOTEBOOK_JSON = {'nbformat': 4, 'metadata': {}}
    PROJECT = MagicMock()
    NOTEBOOK = MagicMock()

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.upload_new_notebook')
    def test_sync_force_new(self, upload_new_notebook_mock):
        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb", new_flag=True)

        # then
        upload_new_notebook_mock.assert_called_with(self.PROJECT, "/tmp/notebook.ipynb", self.NOTEBOOK_JSON)

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.show_unknown_notebook_error')
    def test_sync_force_update_nonexistent_notebook(self, show_unknown_notebook_error_mock):
        # given
        self.PROJECT.get_notebook = MagicMock(side_effect=NotebookNotFound(notebook_id=""))

        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb", update_flag=True)

        # then
        show_unknown_notebook_error_mock.assert_called_with("wajcha/sandbox", "/tmp/notebook.ipynb")

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=UNKNOWN_NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.show_unknown_notebook_error')
    def test_sync_force_update_unknown_notebook(self, show_unknown_notebook_error_mock):
        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb", update_flag=True)

        # then
        show_unknown_notebook_error_mock.assert_called_with("wajcha/sandbox", "/tmp/notebook.ipynb")

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.upload_new_checkpoint')
    def test_sync_force_update(self, upload_new_checkpoint_mock):
        # given
        self.PROJECT.get_notebook = MagicMock(return_value=self.NOTEBOOK)

        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb", update_flag=True)

        # then
        upload_new_checkpoint_mock.assert_called_with(self.PROJECT, self.NOTEBOOK, "/tmp/notebook.ipynb")

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=UNKNOWN_NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.upload_new_notebook')
    def test_sync_unknown_notebook(self, upload_new_notebook_mock):
        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb")

        # then,
        upload_new_notebook_mock.assert_called_with(self.PROJECT, "/tmp/notebook.ipynb", self.UNKNOWN_NOTEBOOK_JSON)

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.upload_new_notebook')
    def test_sync_nonexistent_notebook(self, upload_new_notebook_mock):
        # given
        self.PROJECT.get_notebook = MagicMock(side_effect=NotebookNotFound(notebook_id=""))

        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb")

        # then
        upload_new_notebook_mock.assert_called_with(self.PROJECT, "/tmp/notebook.ipynb", self.NOTEBOOK_JSON)

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.upload_new_checkpoint')
    def test_sync_update_existent_notebook(self, upload_new_checkpoint_mock):
        # given
        self.PROJECT.get_notebook = MagicMock(return_value=self.NOTEBOOK)
        self.NOTEBOOK.get_path = MagicMock(return_value="/tmp/notebook.ipynb")

        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb")

        # then
        upload_new_checkpoint_mock.assert_called_with(self.PROJECT, self.NOTEBOOK, "/tmp/notebook.ipynb")

    @patch('neptune.init', MagicMock(return_value=PROJECT))
    @patch('neptune_notebooks.sync.load_notebook_json', MagicMock(return_value=NOTEBOOK_JSON))
    @patch('neptune_notebooks.sync.verify_file', MagicMock())
    @patch('neptune_notebooks.sync.show_path_changed_error')
    def test_sync_update_existent_notebook_with_new_path(self, show_path_changed_error_mock):
        # given
        self.PROJECT.get_notebook = MagicMock(return_value=self.NOTEBOOK)
        self.NOTEBOOK.get_path = MagicMock(return_value="/tmp/some/other/path/notebook.ipynb")

        # when
        sync("wajcha/sandbox", "/tmp/notebook.ipynb")

        # then
        show_path_changed_error_mock.assert_called_with(
            "wajcha/sandbox",
            "/tmp/some/other/path/notebook.ipynb",
            "/tmp/notebook.ipynb",
            "/tmp/notebook.ipynb")



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import unittest
import random

from click.testing import CliRunner

from neptune_notebooks_plugin import sync


class TestPlugin(unittest.TestCase):
    runner = CliRunner()

    def test_path_not_exist(self):
        path = '/tmp/{}'.format(random.randint(10000, 1000000))
        result = self.runner.invoke(sync, [path])
        self.assertEqual(result.exit_code, 1)
        self.assertEqual(result.output.strip(), "ERROR: File `{}` doesn't exist".format(path))

    def test_path_is_not_file(self):
        path = '/tmp/{}'.format(random.randint(10000, 1000000))
        os.mkdir(path)
        result = self.runner.invoke(sync, [path])
        self.assertEqual(result.exit_code, 1)
        self.assertEqual(result.output.strip(), "ERROR: `{}` is not a file".format(path))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import print_function

# the name of the project
name = 'neptune-notebooks'

# -----------------------------------------------------------------------------
# get on with it
# -----------------------------------------------------------------------------

import os
from os.path import join as pjoin
from glob import glob

from setuptools import setup, find_packages

from setupbase import (create_cmdclass, install_npm, ensure_targets,
                       combine_commands, expand_data_files, set_version_npm)

here = os.path.abspath(os.path.dirname(__file__))

import git_version

version = '0.0.0'
try:
    with open('VERSION') as f:
        version = f.readline().strip()
except IOError:
    pass

nbextension = pjoin(here, 'packages', 'nbextension')
labextension = pjoin(here, 'packages', 'labextension')

# Representative files that should exist after a successful build
jstargets = [
    pjoin(nbextension, 'neptune-notebook.js'),
    pjoin(labextension, 'dist', 'neptune-notebooks-{}.tgz'.format(version)),
]

cmdclass = create_cmdclass(('jsdeps',))

cmdclass['jsdeps'] = combine_commands(
    set_version_npm(path=labextension, version=version, allow_same_version=True),
    install_npm(labextension, build_cmd='dist'),
    ensure_targets(jstargets)
)

cmdclass['git_version'] = git_version.GitVersion

package_data = {
    name: [
        'packages/nbextension/*.*js*',
        'packages/labextension/dist/*.tgz'
    ]
}

data_files = expand_data_files([
    ('share/jupyter/nbextensions/neptune-notebooks', [pjoin(nbextension, '*.js*')]),
    ('share/jupyter/lab/extensions', [pjoin(labextension, 'dist', '*.tgz')]),
])


def main():
    with open(os.path.join(here, 'requirements.txt')) as f:
        requirements = [r.strip() for r in f]
    setup(
        name=name,
        version=version,
        scripts=glob(pjoin('scripts', '*')),
        cmdclass=cmdclass,
        packages=find_packages(here),
        package_data=package_data,
        include_package_data=True,
        data_files=data_files,
        author='Neptune',
        author_email='contact@neptune.ml',
        url='http://jupyter.org',
        license='BSD',
        platforms='any',
        install_requires=requirements,
        keywords=['ipython', 'jupyter'],
        entry_points={
            'neptune.plugins': "notebook = neptune_notebooks_plugin:notebook"
        },
        classifiers=[
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
        ],
    )


if __name__ == "__main__":
    main()
